from dyntastic import Index

from .conftest import MyObject


def test_create_table():
    MyObject.create_table(wait=False)
    assert list(MyObject.scan()) == []


def test_create_table_wait():
    MyObject.create_table()
    assert list(MyObject.scan()) == []


def test_create_table_with_indexes():
    MyObject.create_table(Index("my_bytes"))
    assert list(MyObject.scan()) == []
    assert list(MyObject.scan(index="my_bytes-index")) == []


class MyRegionObject(MyObject):
    __table_region__ = "us-east-2"


def test_create_table_region():
    MyRegionObject._clear_boto3_state()
    assert MyRegionObject._dynamodb_resource().meta.client.meta.region_name == "us-east-2"
    assert MyRegionObject._dynamodb_client().meta.region_name == "us-east-2"
    MyRegionObject.create_table()
    assert list(MyRegionObject.scan()) == []



def test_save(item):
    item.my_str = "new_string"
    assert item.get(item.id, getattr(item, "timestamp", None)).my_str == "foo"
    item.save()
    assert item.my_str == "new_string"
    assert item.get(item.id, getattr(item, "timestamp", None)).my_str == "new_string"


def test_save_different_id(item):
    original_id = (item.id, getattr(item, "timestamp", None))
    item.id = "new_id"
    new_id = ("new_id", original_id[1])

    item.save()

    assert item.get(*new_id) == item
    assert item.get(*original_id).id == original_id[0]

    item.refresh()
    assert item.id == "new_id"


def test_save_aliased_item(alias_item):
    alias_item.save()

    assert alias_item.get(alias_item.id) is not None
    raw_item = alias_item._dynamodb_table().get_item(Key={"id/alias": alias_item.id})["Item"]
    assert raw_item is not None
    assert type(alias_item)(**raw_item).id == alias_item.id



from dyntastic import A

from .conftest import MyObject


def test_refresh_hash(hash_item):
    retrieved = MyObject.get(hash_item.id)
    retrieved.update(A.my_str.set("new_value"))
    assert hash_item.my_str == "foo"
    assert retrieved.my_str == "new_value"

    hash_item.refresh()
    assert hash_item.my_str == "new_value"
    assert retrieved.my_str == "new_value"



import re
from datetime import datetime

import botocore
import pytest

from dyntastic import A
from tests.conftest import MyRangeObject


def test_zero_page_size_errors(populated_model):
    with pytest.raises(
        botocore.exceptions.ParamValidationError, match="valid min value: 1|Invalid range for parameter Limit, value: 0"
    ):
        list(populated_model.query("id1", per_page=0))


@pytest.mark.parametrize("per_page", [1, 2])
def test_query_pager(per_page, populated_model):
    for hash_key in ["id1", "id2", "id3", "id4"]:
        results = list(populated_model.query(hash_key, per_page=per_page))
        assert len(results) == 1
        assert results[0].id == hash_key


@pytest.mark.parametrize("per_page", [1, 2, 3])
def test_query_pager_range(per_page, populated_range_model):
    for hash_key in ["id1", "id2"]:
        results = list(populated_range_model.query(hash_key, per_page=per_page))
        assert len(results) == 2
        assert results[0].id == hash_key
        assert results[1].id == hash_key


def test_query_with_condition(populated_range_model):
    results = list(populated_range_model.query("id1", filter_condition=A.my_int == 1))
    assert len(results) == 1


@pytest.mark.parametrize(
    "range_condition,expected_count",
    [
        (A.timestamp > datetime(2022, 2, 12), 1),
        (A.timestamp >= datetime(2022, 2, 12), 2),
        (A.timestamp > datetime(2022, 2, 13), 0),
        (A.timestamp >= datetime(2022, 2, 13), 1),
        (A.timestamp < datetime(2022, 2, 14), 2),
        (A.timestamp <= datetime(2022, 2, 13), 2),
        (A.timestamp < datetime(2022, 2, 13), 1),
        (A.timestamp <= datetime(2022, 2, 12), 1),
        (A.timestamp < datetime(2022, 2, 12), 0),
        (A.timestamp == datetime(2022, 2, 12), 1),
        (A.timestamp == datetime(2022, 2, 14), 0),
        # note: key expressions do not support !=
        (A.timestamp.begins_with("2022"), 2),
        (A.timestamp.begins_with("2022-02-1"), 2),
        (A.timestamp.begins_with("2022-02-13"), 1),
        (A.timestamp.between("2022", "2023"), 2),
        (A.timestamp.between("2023", "2024"), 0),
        (A.timestamp.between("2022-02-12", "2022-02-12T01"), 1),
    ],
)
def test_query_range_condition(range_condition, expected_count, populated_range_model):
    results = list(populated_range_model.query("id1", range_key_condition=range_condition))
    assert len(results) == expected_count
    assert all(item.id == "id1" for item in results)


@pytest.mark.parametrize("str_value", ["str_1", "str_2"])
@pytest.mark.parametrize("per_page", [1, 2, 3])
def test_query_index_pager(str_value, per_page, populated_model):
    results = list(populated_model.query(A.my_str == str_value, index="my_str-index", per_page=per_page))
    assert len(results) == 2
    assert all(item.my_str == str_value for item in results)


@pytest.mark.parametrize(
    "range_condition,expected_count",
    [
        (A.my_int > 1, 1),
        (A.my_int >= 1, 2),
        (A.my_int > 2, 0),
        (A.my_int >= 2, 1),
        (A.my_int < 3, 2),
        (A.my_int <= 2, 2),
        (A.my_int < 2, 1),
        (A.my_int <= 1, 1),
        (A.my_int < 1, 0),
        (A.my_int == 1, 1),
        (A.my_int == 3, 0),
    ],
)
def test_query_ranged_index_range_condition(range_condition, expected_count, populated_range_model):
    results = list(
        populated_range_model.query(
            A.my_str == "str_1",
            range_key_condition=range_condition,
            index="my_str_my_int-index",
        )
    )
    assert len(results) == expected_count
    assert all(item.my_str == "str_1" for item in results)


@pytest.mark.parametrize("str_value", ["str_1", "str_2"])
@pytest.mark.parametrize("per_page", [1, 2, 3])
def test_query_ranged_index_pager(str_value, per_page, populated_model):
    results = list(populated_model.query(A.my_str == str_value, index="my_str_my_int-index", per_page=per_page))
    assert len(results) == 2
    assert all(item.my_str == str_value for item in results)

    results = list(populated_model.query(A.my_str == str_value, index="my_str_my_int-index", per_page=per_page))
    assert len(results) == 2
    assert all(item.my_str == str_value for item in results)


def test_index_query_must_use_explicit_condition(populated_model):
    with pytest.raises(
        ValueError, match="Must specify attribute condition for index, e.g. A.my_index_hash_key == 'example_value'"
    ):
        list(populated_model.query("str_1", index="my_str_my_int-index"))


def test_cannot_query_consistently_on_index(populated_model):
    with pytest.raises(ValueError, match="Cannot perform a consistent read against a secondary index"):
        list(populated_model.query(A.my_str == "str_1", index="my_str", consistent_read=True))


def test_query_by_page(populated_range_model):
    first_page = populated_range_model.query_page("id1", per_page=1)
    assert len(first_page.items) == 1
    assert first_page.last_evaluated_key == {"id": "id1", "timestamp": "2022-02-12T00:00:00"}
    assert first_page.has_more

    assert str(first_page) == repr(first_page)
    assert "'has_more': True" in str(first_page)
    assert "'last_evaluated_key': {'id': 'id1', 'timestamp': '2022-02-12T00:00:00'}" in str(first_page)

    second_page = populated_range_model.query_page("id1", per_page=1, last_evaluated_key=first_page.last_evaluated_key)
    assert len(second_page.items) == 1
    assert second_page.last_evaluated_key is None
    assert not second_page.has_more


@pytest.mark.parametrize("scan_index_forward", [True, False])
def test_query_scan_index_forward(populated_range_model, scan_index_forward):
    results = list(
        populated_range_model.query(
            "id1", range_key_condition=A.timestamp.begins_with("2022"), scan_index_forward=scan_index_forward
        )
    )

    assert len(results) == 2

    timestamp1 = results[0].timestamp
    timestamp2 = results[1].timestamp

    if scan_index_forward:
        assert timestamp1 < timestamp2
    else:
        assert timestamp2 < timestamp1


@pytest.mark.parametrize("per_page", [25, 100, 2000, None])
def test_query_with_empty_pages_due_to_filter(per_page):
    MyRangeObject.create_table()
    with MyRangeObject.batch_writer():
        for i in range(1000):
            MyRangeObject(
                id="id1", timestamp=datetime(2023, 8, 25, 0, 0, 0, i), my_str="str_1", unindexed_field="unindexed"
            ).save()

    MyRangeObject(
        id="id1", timestamp=datetime(2023, 8, 25, 0, 0, 1), my_str="str_2", unindexed_field="unindexed"
    ).save()

    results = list(
        MyRangeObject.query(
            "id1",
            filter_condition=A.my_str == "str_2",
            per_page=per_page,
        )
    )
    assert len(results) == 1


@pytest.fixture(params=["hash_only", "hash_and_range"])
def keys_only_model(request, populated_model_with_unindexed_field, populated_range_model_with_unindexed_field):
    if request.param == "hash_only":
        return populated_model_with_unindexed_field
    elif request.param == "hash_and_range":
        return populated_range_model_with_unindexed_field


def test_query_manually_refresh_keys_only(keys_only_model):
    result = list(keys_only_model.query(A.my_str == "str_1", index="keys-only-index"))
    assert len(result) == 2

    for item in result:
        assert item.id is not None
        assert item.my_str == "str_1"
        assert item.my_int is not None

        if keys_only_model.__range_key__ is not None:
            assert item.timestamp is not None

        for attr in ("unindexed_field", "my_str_list", "NONEXISTENT"):
            with pytest.raises(
                ValueError,
                match=re.escape(
                    "Dyntastic instance was loaded from a KEYS_ONLY or INCLUDE index. "
                    r"Call refresh() to load the full item, or pass load_full_item=True to query() or scan()"
                ),
            ):
                getattr(item, attr)
                item.unindexed_field

        item.refresh()

        for attr in ("unindexed_field", "my_str_list"):
            getattr(item, attr)


def test_query_auto_refresh_keys_only(keys_only_model):
    result = list(keys_only_model.query(A.my_str == "str_1", index="keys-only-index", load_full_item=True))
    assert len(result) == 2

    for item in result:
        assert item.id is not None
        assert item.my_str == "str_1"
        assert item.my_int is not None

        if keys_only_model.__range_key__ is not None:
            assert item.timestamp is not None

        for attr in ("unindexed_field", "my_str_list"):
            getattr(item, attr)



def test_module_import():
    import dyntastic

    assert dyntastic.__all__ == ["A", "Attr", "DoesNotExist", "Dyntastic", "Index", "transaction", "__version__"]


def test_direct_import():
    from dyntastic import DoesNotExist  # noqa: F401
    from dyntastic import Dyntastic  # noqa: F401
    from dyntastic import __version__  # noqa: F401
    from dyntastic import transaction  # noqa: F401
    from dyntastic import A, Attr, Index  # noqa: F401

    assert A is Attr



import re
from datetime import datetime

import botocore
import pytest

from dyntastic import A


def test_zero_page_size_errors(populated_model):
    with pytest.raises(
        botocore.exceptions.ParamValidationError, match="valid min value: 1|Invalid range for parameter Limit, value: 0"
    ):
        list(populated_model.scan(per_page=0))


@pytest.mark.parametrize("value", [1, 2, 3, 4])
@pytest.mark.parametrize("per_page", [1, 2])
def test_scan_pager(value, per_page, populated_model):
    results = list(populated_model.scan(A.my_int == value, per_page=per_page))
    assert len(results) == 1
    assert results[0].my_int == value


def test_scan_pager_multiple_conditions(populated_range_model):
    results = list(populated_range_model.scan((A.my_int == 1) & (A.my_str == "str_1")))
    assert len(results) == 1
    assert results[0].my_int == 1
    assert results[0].my_str == "str_1"


@pytest.mark.parametrize(
    "condition,expected_count",
    [
        (A.timestamp > datetime(2022, 2, 12), 1),
        (A.timestamp >= datetime(2022, 2, 12), 2),
        (A.timestamp > datetime(2022, 2, 13), 0),
        (A.timestamp >= datetime(2022, 2, 13), 1),
        (A.timestamp < datetime(2022, 2, 14), 2),
        (A.timestamp <= datetime(2022, 2, 13), 2),
        (A.timestamp < datetime(2022, 2, 13), 1),
        (A.timestamp <= datetime(2022, 2, 12), 1),
        (A.timestamp < datetime(2022, 2, 12), 0),
        (A.timestamp == datetime(2022, 2, 12), 1),
        (A.timestamp == datetime(2022, 2, 14), 0),
        (A.timestamp != datetime(2022, 2, 14), 2),
        (A.timestamp != datetime(2022, 2, 12), 1),
        (A.timestamp.begins_with("2022"), 2),
        (A.timestamp.begins_with("2022-02-1"), 2),
        (A.timestamp.begins_with("2022-02-13"), 1),
        (A.timestamp.between("2022", "2023"), 2),
        (A.timestamp.between("2023", "2024"), 0),
        (A.timestamp.between("2022-02-12", "2022-02-12T01"), 1),
    ],
)
def test_scan_range_condition(condition, expected_count, populated_range_model):
    results = list(populated_range_model.scan((A.id == "id1") & condition))
    assert len(results) == expected_count
    assert all(item.id == "id1" for item in results)


@pytest.mark.parametrize("int_values", [(1, 2), (3, 4)])
@pytest.mark.parametrize("per_page", [1, 2, 3])
def test_scan_index_pager(int_values, per_page, populated_model):
    results = list(populated_model.scan(A.my_int.is_in(int_values), index="my_str-index", per_page=per_page))
    assert len(results) == 2
    assert all(item.my_int in int_values for item in results)


def test_scan_by_page(populated_range_model):
    first_page = populated_range_model.scan_page(per_page=2)
    assert len(first_page.items) == 2
    assert first_page.last_evaluated_key == {"id": "id1", "timestamp": "2022-02-13T00:00:00"}
    assert first_page.has_more

    assert str(first_page) == repr(first_page)
    assert "'has_more': True" in str(first_page)
    assert "'last_evaluated_key': {'id': 'id1', 'timestamp': '2022-02-13T00:00:00'}" in str(first_page)

    second_page = populated_range_model.scan_page(per_page=2, last_evaluated_key=first_page.last_evaluated_key)
    assert len(second_page.items) == 2
    assert second_page.last_evaluated_key is None
    assert not second_page.has_more


@pytest.fixture(params=["hash_only", "hash_and_range"])
def keys_only_model(request, populated_model_with_unindexed_field, populated_range_model_with_unindexed_field):
    if request.param == "hash_only":
        return populated_model_with_unindexed_field
    elif request.param == "hash_and_range":
        return populated_range_model_with_unindexed_field


def test_scan_manually_refresh_keys_only(keys_only_model):
    result = list(keys_only_model.scan(index="keys-only-index"))
    assert len(result) == 4

    for item in result:
        assert item.id is not None
        assert item.my_str in ("str_1", "str_2")
        assert item.my_int is not None

        if keys_only_model.__range_key__ is not None:
            assert item.timestamp is not None

        for attr in ("unindexed_field", "my_str_list", "NONEXISTENT"):
            with pytest.raises(
                ValueError,
                match=re.escape(
                    "Dyntastic instance was loaded from a KEYS_ONLY or INCLUDE index. "
                    r"Call refresh() to load the full item, or pass load_full_item=True to query() or scan()"
                ),
            ):
                getattr(item, attr)
                item.unindexed_field

        item.refresh()

        for attr in ("unindexed_field", "my_str_list"):
            getattr(item, attr)


def test_scan_auto_refresh_keys_only(keys_only_model):
    result = list(keys_only_model.scan(index="keys-only-index", load_full_item=True))
    assert len(result) == 4

    for item in result:
        assert item.id is not None
        assert item.my_str in ("str_1", "str_2")
        assert item.my_int is not None

        if keys_only_model.__range_key__ is not None:
            assert item.timestamp is not None

        for attr in ("unindexed_field", "my_str_list"):
            getattr(item, attr)



import pytest
from pydantic import Field

from dyntastic import Dyntastic, pydantic_compat


def test_table_name_required():
    with pytest.raises(ValueError, match="Dyntastic table must have __table_name__ defined"):

        class MyObject(Dyntastic):
            pass

    with pytest.raises(ValueError, match="Dyntastic table must have __table_name__ defined"):

        class MyObjectWithHash(Dyntastic):
            __hash_key__ = "my_hash_key"
            my_hash_key: str

    with pytest.raises(ValueError, match="Dyntastic table must have __table_name__ defined"):

        class MyObjectWithHashAndRange(Dyntastic):
            __hash_key__ = "my_hash_key"
            __range_key__ = "my_range_key"

            my_hash_key: str
            my_range_key: str


def test_hash_key_required():
    with pytest.raises(ValueError, match="Dyntastic table must have __hash_key__ defined"):

        class MyObject(Dyntastic):
            __table_name__ = "my_object"

            my_hash_key: str

    with pytest.raises(ValueError, match="Dyntastic table must have __hash_key__ defined"):

        class MyObjectWithRange(Dyntastic):
            __table_name__ = "my_object"
            __range_key__ = "my_range_key"

            my_range_key: str


def test_key_fields_must_exist():
    with pytest.raises(ValueError, match="Dyntastic __hash_key__ is not defined as a field: 'my_hash_key'"):

        class MyObject(Dyntastic):
            __table_name__ = "my_object"
            __hash_key__ = "my_hash_key"

    with pytest.raises(ValueError, match="Dyntastic __range_key__ is not defined as a field: 'my_range_key'"):

        class MyObjectWithRange(Dyntastic):
            __table_name__ = "my_object"
            __hash_key__ = "my_hash_key"
            __range_key__ = "my_range_key"

            my_hash_key: str


def test_basic_dyntastic_instance():
    class MyObject(Dyntastic):
        __table_name__ = "my_object"
        __hash_key__ = "my_hash_key"

        my_hash_key: str

    instance = MyObject(my_hash_key="example_key")
    assert pydantic_compat.model_dump(instance) == {"my_hash_key": "example_key"}


def test_range_dyntastic_instance():
    class MyObject(Dyntastic):
        __table_name__ = "my_object"
        __hash_key__ = "my_hash_key"
        __range_key__ = "my_range_key"

        my_hash_key: str
        my_range_key: str

    instance = MyObject(my_hash_key="example_key", my_range_key="some_range_key")
    assert pydantic_compat.model_dump(instance) == {"my_hash_key": "example_key", "my_range_key": "some_range_key"}


def test_model_alias():
    class MyAliasObject(Dyntastic):
        __table_name__ = "my_alias_object"
        __hash_key__ = "my/alias"

        my_field: str = Field(..., alias="my/alias")

    instance = MyAliasObject(my_field="test")
    assert pydantic_compat.model_dump(instance, by_alias=True) == {"my/alias": "test"}


def test_table_host():
    class MyObject(Dyntastic):
        __table_name__ = "my_object"
        __hash_key__ = "my_hash_key"
        __table_host__ = "http://localhost:8000"

        my_hash_key: str

    assert MyObject.__table_host__ == "http://localhost:8000"
    client = MyObject._dynamodb_client()
    resource = MyObject._dynamodb_resource()
    assert client.meta.endpoint_url == "http://localhost:8000"
    assert resource.meta.client.meta.endpoint_url == "http://localhost:8000"



from .attr import A, Attr
from .exceptions import DoesNotExist
from .main import Dyntastic, Index, __version__
from .transact import TransactionWriter as transaction

__all__ = ["A", "Attr", "DoesNotExist", "Dyntastic", "Index", "transaction", "__version__"]



import time
import warnings
from decimal import Decimal
from typing import Any, Callable, Dict, Generator, Generic, List, Optional, Tuple, Type, TypeVar, Union

import boto3

try:
    # Python 3.8+
    import importlib.metadata as _metadata
except ModuleNotFoundError:  # pragma: no cover
    # Python 3.7
    import importlib_metadata as _metadata  # type: ignore[no-redef, unused-ignore]

from contextvars import ContextVar

from boto3.dynamodb.conditions import ConditionBase
from pydantic import BaseModel, PrivateAttr

from . import attr, pydantic_compat, transact
from .attr import Attr, _UpdateAction, translate_updates
from .batch import BatchWriter, invoke_with_backoff
from .exceptions import DoesNotExist
from .transact import current_transaction_writer

__version__ = _metadata.version("dyntastic")


_T = TypeVar("_T", bound="Dyntastic")


class _TableMetadata:
    __table_name__: Union[str, Callable[[], str]]
    __table_region__: Optional[str] = None
    __table_host__: Optional[str] = None

    __hash_key__: str
    __range_key__: Optional[str] = None

    _dyntastic_batch_writer: ContextVar[Optional[BatchWriter]]


class ResultPage(Generic[_T]):
    def __init__(self, items: List[_T], last_evaluated_key: Optional[dict]):
        self.items = items
        self.last_evaluated_key = last_evaluated_key
        self.has_more = last_evaluated_key is not None

    def __str__(self):
        return f"ResultPage: {self.__dict__}"

    def __repr__(self):
        return str(self)


class Index:
    def __init__(
        self,
        hash_key: str,
        range_key: Optional[str] = None,
        index_name: Optional[str] = None,
        keys_only: bool = False,
    ):
        self.hash_key = hash_key
        self.range_key = range_key
        # TODO: support INCLUDE projection?
        self.projection = "KEYS_ONLY" if keys_only else "ALL"

        if not index_name:
            if range_key:
                index_name = f"{hash_key}_{range_key}-index"
            else:
                index_name = f"{hash_key}-index"

        self.index_name = index_name


class Dyntastic(_TableMetadata, pydantic_compat.BaseModel):
    _dyntastic_unrefreshed: bool = PrivateAttr(default=False)
    _dyntastic_missing_attributes_from_index: bool = PrivateAttr(default=False)

    @classmethod
    def get_model(cls, item: dict):
        """Get a model instance from a DynamoDB item.

        This method can be overridden to support a single-table design pattern
        (i.e. multiple schemas shared in a single table).
        """

        return cls

    @classmethod
    def _dyntastic_load_model(cls, item: dict, load_full_item: bool = False):
        model = cls.get_model(item)

        data, had_validation_errors = pydantic_compat.try_model_construct(model, item)
        if had_validation_errors:
            # assume KEYS_ONLY or INCLUDE index
            data._dyntastic_missing_attributes_from_index = True

        if load_full_item:
            data.refresh()

        return data

    @classmethod
    def get(cls: Type[_T], hash_key, range_key=None, *, consistent_read: bool = False) -> _T:
        if cls.__range_key__ and range_key is None:
            raise ValueError(f"Must provide range_key to {cls.__name__}.get()")
        elif range_key and cls.__range_key__ is None:
            raise ValueError(f"Did not expect range_key for {cls.__name__}.get(), found '{range_key}'")

        key = {cls.__hash_key__: hash_key}
        if cls.__range_key__:
            key[cls.__range_key__] = range_key

        serialized_key = attr.serialize(key)

        response = cls._dynamodb_table().get_item(Key=serialized_key, ConsistentRead=consistent_read)
        data = response.get("Item")
        if data:
            return cls._dyntastic_load_model(data)
        else:
            raise DoesNotExist

    @classmethod
    def safe_get(cls: Type[_T], hash_key, range_key=None, *, consistent_read: bool = False) -> Optional[_T]:
        try:
            return cls.get(hash_key, range_key=range_key, consistent_read=consistent_read)
        except DoesNotExist:
            return None

    @classmethod
    def batch_get(
        cls: Type[_T],
        keys: Union[List[str], List[Tuple[str, str]]],
        consistent_read: bool = False,
    ) -> List[_T]:
        hash_key_type = pydantic_compat.field_type(cls, cls.__hash_key__)

        if cls.__range_key__ and not all(isinstance(key, (list, tuple)) and len(key) == 2 for key in keys):
            raise ValueError(f"Must provide (hash_key, range_key) tuples as `keys` to {cls.__name__}.batch_get()")

        if cls.__range_key__ is None and not all(isinstance(key, hash_key_type) for key in keys):
            raise ValueError(
                f"Must only provide {hash_key_type.__name__} types as `keys` to {cls.__name__}.batch_get()"
            )

        serialized_keys = []
        for key in keys:
            if cls.__range_key__:
                key_dict = {cls.__hash_key__: key[0], cls.__range_key__: key[1]}
            else:
                assert isinstance(key, hash_key_type)
                key_dict = {cls.__hash_key__: key}

            serialized_keys.append(attr.serialize(key_dict))

        responses = invoke_with_backoff(
            cls._dynamodb_resource().batch_get_item,
            {cls._resolve_table_name(): {"Keys": serialized_keys, "ConsistentRead": consistent_read}},
            "UnprocessedKeys",
        )

        items: List[_T] = []
        for response in responses:
            raw_items = response["Responses"][cls._resolve_table_name()]
            items.extend(cls._dyntastic_load_model(item) for item in raw_items)

        return items

    @classmethod
    def query(
        cls: Type[_T],
        hash_key,
        *,
        consistent_read: bool = False,
        range_key_condition=None,
        filter_condition: Optional[ConditionBase] = None,
        index: Optional[str] = None,
        per_page: Optional[int] = None,
        last_evaluated_key: Optional[dict] = None,
        scan_index_forward: bool = True,
        load_full_item: bool = False,
    ) -> Generator[_T, None, None]:
        while True:
            result = cls.query_page(
                hash_key,
                consistent_read=consistent_read,
                range_key_condition=range_key_condition,
                filter_condition=filter_condition,
                index=index,
                per_page=per_page,
                last_evaluated_key=last_evaluated_key,
                scan_index_forward=scan_index_forward,
                load_full_item=load_full_item,
            )

            last_evaluated_key = result.last_evaluated_key
            yield from result.items

            if not result.has_more:
                break  # pragma: no cover (in python 3.8/3.9, this appeared as missing coverage)

    @classmethod
    def query_page(
        cls: Type[_T],
        hash_key: Union[str, ConditionBase],
        *,
        consistent_read: bool = False,
        range_key_condition: Optional[ConditionBase] = None,
        filter_condition: Optional[ConditionBase] = None,
        index: Optional[str] = None,
        per_page: Optional[int] = None,
        last_evaluated_key: Optional[dict] = None,
        scan_index_forward: bool = True,
        load_full_item: bool = False,
    ) -> ResultPage[_T]:
        if index and consistent_read:
            raise ValueError("Cannot perform a consistent read against a secondary index")

        if isinstance(hash_key, ConditionBase):
            key_condition = hash_key
        elif index is not None:
            raise ValueError("Must specify attribute condition for index, e.g. A.my_index_hash_key == 'example_value'")
        else:
            key_condition: ConditionBase = Attr(cls.__hash_key__) == hash_key  # type: ignore

        if range_key_condition:
            key_condition &= range_key_condition

        response = cls._dyntastic_call(
            "query",
            ConsistentRead=consistent_read,
            IndexName=index,
            Limit=per_page,
            ExclusiveStartKey=last_evaluated_key,
            KeyConditionExpression=key_condition,
            FilterExpression=filter_condition,
            ScanIndexForward=scan_index_forward,
        )

        raw_items = response.get("Items")
        items = [cls._dyntastic_load_model(item, load_full_item=load_full_item) for item in raw_items]
        last_evaluated_key = response.get("LastEvaluatedKey")

        return ResultPage(items, last_evaluated_key)

    @classmethod
    def scan(
        cls: Type[_T],
        filter_condition: Optional[ConditionBase] = None,
        *,
        consistent_read: bool = False,
        index: Optional[str] = None,
        per_page: Optional[int] = None,
        last_evaluated_key: Optional[dict] = None,
        load_full_item: bool = False,
    ):
        while True:
            result = cls.scan_page(
                filter_condition=filter_condition,
                consistent_read=consistent_read,
                index=index,
                per_page=per_page,
                last_evaluated_key=last_evaluated_key,
                load_full_item=load_full_item,
            )

            last_evaluated_key = result.last_evaluated_key
            yield from result.items

            if not result.has_more:
                break

    @classmethod
    def scan_page(
        cls: Type[_T],
        filter_condition: Optional[ConditionBase] = None,
        *,
        consistent_read: bool = False,
        index: Optional[str] = None,
        per_page: Optional[int] = None,
        last_evaluated_key: Optional[dict] = None,
        load_full_item: bool = False,
    ) -> ResultPage[_T]:
        response = cls._dyntastic_call(
            "scan",
            ConsistentRead=consistent_read,
            IndexName=index,
            Limit=per_page,
            ExclusiveStartKey=last_evaluated_key,
            FilterExpression=filter_condition,
        )

        raw_items = response.get("Items")
        items = [cls._dyntastic_load_model(item, load_full_item=load_full_item) for item in raw_items]
        last_evaluated_key = response.get("LastEvaluatedKey")

        return ResultPage(items, last_evaluated_key)

    def save(self, *, condition: Optional[ConditionBase] = None):
        data = pydantic_compat.model_dump(self, by_alias=True)
        dynamo_serialized = attr.serialize(data)
        return self._dyntastic_call("put_item", Item=dynamo_serialized, ConditionExpression=condition)

    def delete(self, *, condition: Optional[ConditionBase] = None):
        return self._dyntastic_call("delete_item", Key=self._dyntastic_key_dict, ConditionExpression=condition)

    # TODO: Support ReturnValues
    def update(
        self,
        *actions: _UpdateAction,
        condition: Optional[ConditionBase] = None,
        require_condition: bool = False,
        refresh: bool = True,
    ):
        if not actions:
            raise ValueError("Must provide at least one action to update")

        # TODO: Run all of the expression value through pydantic validators on
        # the class, to support all of the various input type casting (do this
        # before serialize)
        update_data: Dict[str, Any] = attr.serialize(translate_updates(*actions))
        try:
            response = self._dyntastic_call(
                "update_item",
                Key=self._dyntastic_key_dict,
                ConditionExpression=condition,
                **update_data,
            )
            self._dyntastic_unrefreshed = True
            if refresh:
                if current_transaction_writer() is not None:
                    warnings.warn("Cannot refresh model in transaction, skipping refresh", stacklevel=2)
                else:
                    # TODO: utilize ReturnValues in response when possible
                    self.refresh()

            return response
        except self.ConditionException():
            if require_condition:
                raise

    def refresh(self):
        self._dyntastic_unrefreshed = False
        self._dyntastic_missing_attributes_from_index = False
        data = self.get(self._dyntastic_hash_key, self._dyntastic_range_key)
        self.__dict__.update(data.__dict__)

    def transaction_condition(self, condition: ConditionBase):
        transaction_writer = current_transaction_writer()
        if transaction_writer is None:
            raise Exception(f"Cannot use {self.__class__.__name__}.transaction_condition() outside of a transaction")

        item = self._construct_transact_item(
            "transaction_condition",
            {"Key": self._dyntastic_key_dict, "ConditionExpression": condition},
        )
        transaction_writer.add(self.__class__, item)

    @classmethod
    def batch_writer(cls, batch_size: int = 25):
        return BatchWriter(cls, batch_size=batch_size)

    @classmethod
    def submit_batch_write(cls, batch: List[dict]):
        if not batch:
            return

        responses = invoke_with_backoff(
            cls._dynamodb_resource().batch_write_item,
            {cls._resolve_table_name(): batch},
            "UnprocessedItems",
        )

        return responses

    # Note: This cannot use @classmethod and @property together for python <3.9
    @classmethod
    def ConditionException(cls):
        return cls._dynamodb_table().meta.client.exceptions.ConditionalCheckFailedException

    # TODO: support more configuration for new table
    @classmethod
    def create_table(cls, *indexes: Union[str, Index], wait: bool = True):
        """Creates a DynamoDB table (primarily for testing, limited configuration supported)"""

        throughput = {"ReadCapacityUnits": 1, "WriteCapacityUnits": 1}
        attributes = {cls.__hash_key__}
        key_schema = [{"AttributeName": cls.__hash_key__, "KeyType": "HASH"}]
        if cls.__range_key__:
            attributes.add(cls.__range_key__)
            key_schema.append({"AttributeName": cls.__range_key__, "KeyType": "RANGE"})

        kwargs = {}
        if indexes:
            secondary_indexes = []
            for index in indexes:
                if isinstance(index, str):
                    index = Index(index)

                attributes.add(index.hash_key)
                index_schema = [{"AttributeName": index.hash_key, "KeyType": "HASH"}]
                if index.range_key:
                    attributes.add(index.range_key)
                    index_schema.append({"AttributeName": index.range_key, "KeyType": "RANGE"})

                secondary_indexes.append(
                    {
                        "IndexName": index.index_name,
                        "KeySchema": index_schema,
                        "Projection": {"ProjectionType": index.projection},
                        "ProvisionedThroughput": {"ReadCapacityUnits": 1, "WriteCapacityUnits": 1},
                    }
                )

            kwargs["GlobalSecondaryIndexes"] = secondary_indexes

        attribute_definitions = [
            {"AttributeName": attr, "AttributeType": cls._dynamodb_type(attr)} for attr in attributes
        ]

        cls._dynamodb_resource().create_table(
            TableName=cls._resolve_table_name(),
            KeySchema=key_schema,
            AttributeDefinitions=attribute_definitions,
            ProvisionedThroughput=throughput,
            **kwargs,
        )

        if wait:
            cls._wait_until_exists()

    # Internal helpers

    @classmethod
    def _resolve_table_name(cls) -> str:
        if callable(cls.__table_name__):
            return cls.__table_name__()
        else:
            return cls.__table_name__

    @classmethod
    def _dynamodb_type(cls, key: str) -> str:
        # Note: pragma nocover on the following line as coverage marks the ->exit branch as
        # being missed (since we can always find a field matching the key passed in)
        python_type = next(
            pydantic_compat.annotation(field)
            for field_name, field in pydantic_compat.model_fields(cls).items()
            if pydantic_compat.alias(field_name, field) == key
        )  # pragma: nocover
        if python_type == bytes:
            return "B"
        elif python_type in (int, Decimal, float):
            return "N"
        else:
            # TODO: how to properly differentiate between types like datetime
            # which serialize to str, and other types that do not?
            # TODO: use boto3.dynamodb.types.TypeSerializer._get_dynamodb_type() as a reference
            return "S"

    @property
    def _dyntastic_hash_key(self):
        return getattr(self, self.__hash_key__)

    @property
    def _dyntastic_range_key(self):
        if self.__range_key__:
            return getattr(self, self.__range_key__)
        else:
            return None

    @property
    def _dyntastic_key_dict(self):
        key = {self.__hash_key__: self._dyntastic_hash_key}
        if self.__range_key__:
            key[self.__range_key__] = self._dyntastic_range_key

        return attr.serialize(key)

    @classmethod
    def _dynamodb_boto3_kwargs(cls):
        kwargs = {}

        if cls.__table_region__:
            kwargs["region_name"] = cls.__table_region__

        if cls.__table_host__:
            kwargs["endpoint_url"] = cls.__table_host__

        return kwargs

    @classmethod
    def _dynamodb_resource(cls):
        if cls._dynamodb_resource_instance is None:  # type: ignore
            kwargs = cls._dynamodb_boto3_kwargs()
            cls._dynamodb_resource_instance = boto3.resource("dynamodb", **kwargs)  # type: ignore
        return cls._dynamodb_resource_instance  # type: ignore

    @classmethod
    def _dynamodb_table(cls):
        if cls._dynamodb_table_instance is None:  # type: ignore
            cls._dynamodb_table_instance = cls._dynamodb_resource().Table(cls._resolve_table_name())  # type: ignore
        return cls._dynamodb_table_instance  # type: ignore

    @classmethod
    def _dynamodb_client(cls):
        if cls._dynamodb_client_instance is None:  # type: ignore
            kwargs = cls._dynamodb_boto3_kwargs()
            cls._dynamodb_client_instance = boto3.client("dynamodb", **kwargs)  # type: ignore
        return cls._dynamodb_client_instance  # type: ignore

    @classmethod
    def _wait_until_exists(cls):
        # wait a maximum of 15 * 2 = 30 seconds
        for _ in range(15):  # pragma: no cover
            response = cls._dynamodb_client().describe_table(TableName=cls._resolve_table_name())
            if response["Table"].get("TableStatus") == "ACTIVE":  # pragma: no cover
                break

            time.sleep(2)

    @classmethod
    def _clear_boto3_state(cls):
        cls._dynamodb_table_instance = None  # type: ignore
        cls._dynamodb_resource_instance = None  # type: ignore
        cls._dynamodb_client_instance = None  # type: ignore

    @classmethod
    def _construct_batch_item(cls, operation: str, filtered_kwargs: Dict[str, Any]):
        if operation == "delete_item":
            method = "delete"
            key = "DeleteRequest"
            required_kwargs = {"Key"}
        elif operation == "put_item":
            method = "save"
            key = "PutRequest"
            required_kwargs = {"Item"}
        else:  # pragma: nocover
            raise ValueError(f"Operation {operation} not supported with {cls.__name__}.batch_writer()")

        if filtered_kwargs.keys() != required_kwargs:
            raise ValueError(
                f"Cannot provide additional arguments to {cls.__name__}.{method}() when using batch_writer()"
            )

        return {key: filtered_kwargs}

    @classmethod
    def _construct_transact_item(cls, operation: str, filtered_kwargs: Dict[str, Any]):
        filtered_kwargs["TableName"] = cls._resolve_table_name()

        if "ConditionExpression" in filtered_kwargs:
            condition_data = transact.serialize_condition(filtered_kwargs["ConditionExpression"])
            filtered_kwargs["ConditionExpression"] = condition_data["ConditionExpression"]

            # Merging condition expression and update expression names/values so they are both present.
            # boto3 names/values look like '#n...' and ':v...', while dyntastic uses just '#...' and ':...'
            # so they should be mutually exclusive and not overlap at all
            names = filtered_kwargs.setdefault("ExpressionAttributeNames", {})
            names.update(condition_data["ExpressionAttributeNames"])

            values = filtered_kwargs.setdefault("ExpressionAttributeValues", {})
            values.update(condition_data["ExpressionAttributeValues"])

        for data_key in ("Key", "Item", "ExpressionAttributeValues"):
            if data_key in filtered_kwargs:
                filtered_kwargs[data_key] = transact.serialize_data(filtered_kwargs[data_key])

        key = {
            "delete_item": "Delete",
            "put_item": "Put",
            "update_item": "Update",
            "transaction_condition": "ConditionCheck",
        }.get(operation)

        if key is None:  # pragma: nocover
            raise ValueError(f"Operation {operation} not supported with dyntastic.TransactionWriter")

        return {key: filtered_kwargs}

    @classmethod
    def _dyntastic_call(cls, operation: str, **kwargs):
        method = getattr(cls._dynamodb_table(), operation)
        filtered_kwargs = {key: value for key, value in kwargs.items() if value is not None}

        batch_writer = cls._dyntastic_batch_writer.get()
        transaction_writer = current_transaction_writer()

        if batch_writer is not None and transaction_writer is not None:
            raise ValueError("Cannot use batch_writer() and transaction() at the same time")

        if (batch_writer is None and transaction_writer is None) or operation in ["query", "scan"]:
            return method(**filtered_kwargs)

        if batch_writer is not None:
            batch_item = cls._construct_batch_item(operation, filtered_kwargs)
            batch_writer.add(batch_item)
        elif transaction_writer is not None:
            item = cls._construct_transact_item(operation, filtered_kwargs)
            transaction_writer.add(cls, item)
        else:  # pragma: nocover
            raise Exception("Logically will always have a batch or transaction writer here")

    def ignore_unrefreshed(self):
        self._dyntastic_unrefreshed = False

    def _get_private_field(self, attr: str):
        try:
            return getattr(self, attr)
        except AttributeError:  # pragma: nocover
            # Note: Without this remapping AttributeError -> Exception, it is
            # particularly difficult to debug the issues that arise. For
            # example, without "model_post_init" in the __getattribute__
            # function below, the error appears as all model fields raising
            # AttributeError on access due to private fields like
            # _dyntastic_unrefreshed triggering that during pydantic's
            # __getattr__.
            #
            # Long story short, this should catch bugs in a much more easy-to-debug way.

            raise Exception(f"{attr} could not be accessed, dyntastic<->pydantic bug")

    def __getattribute__(self, attr: str):
        # breakpoint()
        # All of the code in this function works to "disable" an instance
        # that has been updated with refresh=False, to avoid accidentally
        # working with stale data

        if attr.startswith("_") or attr in {
            "refresh",
            "ignore_unrefreshed",
            "ConditionException",
            # Note: Without model_post_init here, _dyntastic_unrefreshed will
            # be accessed below before pydantic v2 is fully initialized,
            # which causes a bad state (for example, no field attribute can be accessed on the class)
            "model_post_init",
        }:
            return super().__getattribute__(attr)

        if self._get_private_field("_dyntastic_unrefreshed"):
            raise ValueError(
                "Dyntastic instance was not refreshed after update. "
                "Call refresh(), or use ignore_unrefreshed() to ignore safety checks"
            )

        try:
            return super().__getattribute__(attr)
        except AttributeError:
            if self._get_private_field("_dyntastic_missing_attributes_from_index"):
                raise ValueError(
                    "Dyntastic instance was loaded from a KEYS_ONLY or INCLUDE index. "
                    "Call refresh() to load the full item, or pass load_full_item=True to query() or scan()"
                )
            raise

    def __init_subclass__(cls, **kwargs):
        # Note: in pydantic v2, our private attributes like __hash_key__ are
        # not exposed on the model until the class is fully initialized, at
        # which point __pydantic_init_subclass__ is called.
        if pydantic_compat.IS_VERSION_1:  # pragma: nocover
            cls.__pydantic_init_subclass__(**kwargs)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs):
        if not pydantic_compat.IS_VERSION_1:  # pragma: nocover
            super().__pydantic_init_subclass__(**kwargs)  # type: ignore[unused-ignore, misc]

        cls._clear_boto3_state()

        cls._dyntastic_batch_writer = ContextVar("dyntastic_batch_writer", default=None)

        if not hasattr(cls, "__table_name__"):
            raise ValueError("Dyntastic table must have __table_name__ defined")

        if not hasattr(cls, "__hash_key__"):
            raise ValueError("Dyntastic table must have __hash_key__ defined")

        if not _has_alias(cls, cls.__hash_key__):
            raise ValueError(f"Dyntastic __hash_key__ is not defined as a field: '{cls.__hash_key__}'")

        if cls.__range_key__ and not _has_alias(cls, cls.__range_key__):
            raise ValueError(f"Dyntastic __range_key__ is not defined as a field: '{cls.__range_key__}'")


def _has_alias(model: Type[BaseModel], name: str) -> bool:
    for field_name, field in pydantic_compat.model_fields(model).items():
        if pydantic_compat.alias(field_name, field) == name:
            return True

    return False



import pytest

from .conftest import MyIntObject, MyObject, MyRangeObject, query_data, range_query_data

hash_keys = [item["id"] for item in query_data]
loaded_hash_data = [MyObject(**item) for item in query_data]

hash_range_keys = [(item["id"], item["timestamp"]) for item in range_query_data]
loaded_range_data = [MyRangeObject(**item) for item in range_query_data]


def test_get_by_hash_key(populated_model):
    assert populated_model.batch_get(hash_keys) == loaded_hash_data
    assert populated_model.batch_get([*hash_keys, "nonexistent"]) == loaded_hash_data


def test_get_by_int_hash_key(populated_int_model):
    assert populated_int_model.batch_get([1, 2, 3]) == [MyIntObject(id=1), MyIntObject(id=2), MyIntObject(id=3)]
    assert populated_int_model.batch_get([1, 2, 3, 1000]) == [MyIntObject(id=1), MyIntObject(id=2), MyIntObject(id=3)]


def test_get_by_hash_key_and_range_key(populated_range_model):
    assert populated_range_model.batch_get(hash_range_keys) == loaded_range_data
    assert populated_range_model.batch_get([*hash_range_keys, ("nonexistent", "TS")]) == loaded_range_data


def test_invalid_keys(populated_model):
    error_message = rf"Must only provide str types as `keys` to {populated_model.__name__}\.batch_get\(\)"

    with pytest.raises(ValueError, match=error_message):
        assert populated_model.batch_get([("hash", "range")]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_model.batch_get(["good", ("hash", "range")]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_model.batch_get(["good", ("hash", "range"), "good2"]) == []


def test_invalid_int_keys(populated_int_model):
    error_message = rf"Must only provide int types as `keys` to {populated_int_model.__name__}\.batch_get\(\)"

    with pytest.raises(ValueError, match=error_message):
        assert populated_int_model.batch_get([1, "bad"]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_int_model.batch_get([1, "bad", 2]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_int_model.batch_get([1, (2, "bad_range")]) == []


def test_invalid_keys_on_range_model(populated_range_model):
    error_message = (
        rf"Must provide \(hash_key, range_key\) tuples as `keys` to {populated_range_model.__name__}\.batch_get\(\)"
    )

    with pytest.raises(ValueError, match=error_message):
        assert populated_range_model.batch_get(["hash"]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_range_model.batch_get([("good", "range"), "hash"]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_range_model.batch_get([("missing_range",)]) == []

    with pytest.raises(ValueError, match=error_message):
        assert populated_range_model.batch_get([("hash", "range", "extra")]) == []



from pathlib import Path

from setuptools import find_packages, setup

ROOT_DIRECTORY = Path(__file__).resolve().parent

readme = (ROOT_DIRECTORY / "README.md").read_text()
description = next(
    line for line in readme.splitlines() if line and not line.startswith("#") and not line.startswith("[")
)
changelog = (ROOT_DIRECTORY / "CHANGELOG.md").read_text()
long_description = readme + "\n\n" + changelog


DEV_REQUIRES = [
    "black",
    "boto3-stubs[dynamodb]",
    "coverage",
    "flake8==6.0.0; python_version > '3.7'",
    "flake8<5; python_version == '3.7'",
    "flake8-bugbear",
    "isort",
    "moto==4.2.2",
    "mypy",
    "pre-commit",
    "pytest",
    "pytest-cov",
    "pytest-mock",
    "pytest-sugar",
    "pytest-xdist",
]

setup(
    name="dyntastic",
    version="0.13.0",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
    author="Naya Verdier",
    url="https://github.com/nayaverdier/dyntastic",
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "boto3>=1.10.0; python_version < '3.12'",
        "boto3>=1.20.0; python_version >= '3.12'",
        "pydantic>=1.7.1,<3",
        "importlib-metadata>=1.0.0; python_version < '3.8'",
    ],
    python_requires=">=3.7",
    extras_require={
        "dev": DEV_REQUIRES,
        "deploy": ["twine"],
    },
    package_data={"dyntastic": ["py.typed"]},
    include_package_data=True,
)



import time
from contextvars import Token
from typing import Callable, Optional, Type

from . import main


def invoke_with_backoff(
    f: Callable,
    request_items: dict,
    unprocessed_key: str,
    max_attempts: int = 5,
):
    responses = []
    backoff_sleep = 0.05
    attempts = 0

    unprocessed_items = request_items
    while unprocessed_items:
        response = f(RequestItems=unprocessed_items)
        responses.append(response)

        unprocessed_items = response[unprocessed_key]

        if unprocessed_items:  # pragma: no cover
            attempts += 1
            if attempts > max_attempts:
                raise Exception(f"Exceeded max attempts ({max_attempts}) to process unprocessed keys")

            time.sleep(backoff_sleep)
            backoff_sleep *= 2

    return responses


class BatchWriter:
    def __init__(self, table: Type["main.Dyntastic"], batch_size: int = 25):
        self.table = table
        self.batch_size = batch_size
        self.batch: list = []
        self.batches_submitted = 0
        self._context_var_reset_token: Optional[Token] = None

    def __enter__(self):
        self._context_var_reset_token = self.table._dyntastic_batch_writer.set(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        assert self._context_var_reset_token is not None
        self.table._dyntastic_batch_writer.reset(self._context_var_reset_token)
        self._context_var_reset_token = None

        if not exc_type:
            self._commit()

    def _commit(self):
        if self.batch:
            self.table.submit_batch_write(self.batch)
            self.batch = []
            self.batches_submitted += 1

    def add(self, item):
        self.batch.append(item)
        if len(self.batch) >= self.batch_size:
            self._commit()



import pytest

from dyntastic import DoesNotExist

from .conftest import MyIntObject, MyObject, MyRangeObject


def test_get_by_hash_key(hash_item):
    retrieved = MyObject.get(hash_item.id)
    safe_retrieved = MyObject.safe_get(hash_item.id)
    assert retrieved == safe_retrieved == hash_item


def test_get_by_int_hash_key(populated_int_model):
    retrieved = MyIntObject.get(1)
    safe_retrieved = MyIntObject.safe_get(1)
    assert retrieved == safe_retrieved == MyIntObject(id=1)


def test_get_with_range_key(range_item):
    retrieved = MyRangeObject.get(range_item.id, range_item.timestamp)
    safe_retrieved = MyRangeObject.safe_get(range_item.id, range_item.timestamp)
    assert retrieved == safe_retrieved == range_item


def test_range_key_required(range_item):
    with pytest.raises(ValueError, match="Must provide range_key to MyRangeObject.get\\(\\)"):
        MyRangeObject.get(range_item.id)


def test_range_key_not_expected(hash_item):
    with pytest.raises(ValueError, match="Did not expect range_key for MyObject.get\\(\\), found 'my_range_value'"):
        MyObject.get(hash_item.id, "my_range_value")


def test_get_nonexistent(hash_item, range_item):
    with pytest.raises(DoesNotExist):
        MyObject.get("nonexistent")

    with pytest.raises(ValueError, match="Must provide range_key to MyRangeObject.get\\(\\)"):
        MyRangeObject.get("nonexistent")

    with pytest.raises(DoesNotExist):
        MyRangeObject.get("nonexistent", "nonexistent_range")

    assert MyObject.safe_get("nonexistent") is None
    assert MyRangeObject.safe_get("nonexistent", "nonexistent_range") is None

    with pytest.raises(ValueError, match="Must provide range_key to MyRangeObject.get\\(\\)"):
        MyRangeObject.safe_get("nonexistent")


def test_get_nonexistent_int_hash_key(populated_int_model):
    with pytest.raises(DoesNotExist):
        MyIntObject.get(100)

    assert MyIntObject.safe_get(100) is None



import re

import pytest

from dyntastic import A
from tests.conftest import MyObject


def test_batch_write():
    MyObject.create_table()

    with MyObject.batch_writer() as writer:
        MyObject(id="1").save()
        assert list(MyObject.scan()) == []
        second_object = MyObject(id="2")
        second_object.save()
        assert list(MyObject.scan()) == []
        MyObject(id="3").save()
        assert list(MyObject.scan()) == []

    assert writer.batches_submitted == 1

    assert len(list(MyObject.scan())) == 3
    MyObject(id="4").save()
    assert len(list(MyObject.scan())) == 4

    with MyObject.batch_writer():
        MyObject(id="5").save()
        MyObject(id="1").delete()
        second_object.delete()

    assert len(list(MyObject.scan())) == 3


def test_batch_write_with_batch_size():
    MyObject.create_table()

    with MyObject.batch_writer(batch_size=2):
        MyObject(id="1").save()
        assert list(MyObject.scan()) == []
        MyObject(id="2").save()
        assert len(list(MyObject.scan())) == 2
        MyObject(id="3").save()
        assert len(list(MyObject.scan())) == 2

    assert len(list(MyObject.scan())) == 3


def test_batch_write_with_batch_size_no_exit_submit():
    MyObject.create_table()

    with MyObject.batch_writer(batch_size=2):
        MyObject(id="1").save()
        assert list(MyObject.scan()) == []
        MyObject(id="7").save()
        assert len(list(MyObject.scan())) == 2
        MyObject(id="3").save()
        assert len(list(MyObject.scan())) == 2
        MyObject(id="4").save()
        assert len(list(MyObject.scan())) == 4

    # add coverage for skipping submit when no items in batch
    MyObject.submit_batch_write([])

    assert len(list(MyObject.scan())) == 4


def test_batch_write_save_with_conditions_errors():
    MyObject.create_table()

    with pytest.raises(
        ValueError,
        match=re.escape("Cannot provide additional arguments to MyObject.save() when using batch_writer()"),
    ):
        with MyObject.batch_writer():
            MyObject(id="1").save(condition=A.id == "1")


def test_batch_write_delete_with_conditions_errors():
    MyObject.create_table()

    with pytest.raises(
        ValueError,
        match=re.escape("Cannot provide additional arguments to MyObject.delete() when using batch_writer()"),
    ):
        with MyObject.batch_writer():
            MyObject(id="1").delete(condition=A.id == "1")
